import matplotlib.pyplot as plt
import mglearn

X, y = mglearn.datasets.make_forge()
mglearn.discrete_scatter(X[:, 0], X[:, 1], y)
plt.legend(["Class 0", "Class 1"], loc=4)
plt.xlabel("First feature")
plt.ylabel("Second feature")
print("X.shape: {}".format(X.shape))
#plt.show()


mglearn.plots.plot_knn_classification(n_neighbors=1)


import statistics as stat

from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder


def separate_target(data, colName):
    # Separate target from predictors
    data.dropna(axis=0, subset=[colName], inplace=True)
    y = data[colName]
    X = data.drop([colName], axis=1)
    return X, y


def split_columns(x):
    # Select Categorical & numerical columns
    categorical = [cname for cname in x.columns if x[cname].nunique() < 10 and
                   x[cname].dtype == "object"]
    numerical = [cname for cname in x.columns if x[cname].dtype in ['int64', 'float64']]

    return categorical, numerical


def split_train_test_80(x_train, y_train):
    return train_test_split(x_train, y_train, train_size=0.8, test_size=0.2,
                            random_state=0)


def preprocess(X):
    categorical_cols, numerical_cols = split_columns(X)
    my_cols = categorical_cols + numerical_cols
    numerical_transformer = SimpleImputer(strategy='constant')
    categorical_transformer = Pipeline(steps=[
        ('imputer', SimpleImputer(strategy='most_frequent')),
        ('onehot', OneHotEncoder(handle_unknown='ignore'))
    ])
    X_train_full = X[my_cols].copy()

    return ColumnTransformer(
        transformers=[
            ('num', numerical_transformer, numerical_cols),
            ('cat', categorical_transformer, categorical_cols)
        ]), X_train_full


def get_scores(model, preprocessor, X_train, y_train, cv):
    # - Pipeline
    model_pipeline = Pipeline(steps=[('preprocessor', preprocessor),
                                     ('model', model)
                                     ])

    scores = -1 * cross_val_score(model_pipeline, X_train, y_train,
                                  cv=cv,
                                  scoring='neg_mean_absolute_error')

    return stat.median_grouped(scores, interval=1)



import keras_tuner


def hyperband_search(regressor_builder, X_train, y_train, X_valid, y_valid, max_epochs=5, batch_size=128):
    # tuner = keras_tuner.RandomSearch(
    tuner = keras_tuner.Hyperband(
        hypermodel=regressor_builder,
        # The objective name and direction.
        # Name is the f"val_{snake_case_metric_class_name}".
        objective=keras_tuner.Objective("val_mean_absolute_error", direction="min"),
        max_epochs=50,
        factor=3,
        seed=42,
        # max_trials=2,
        overwrite=True,
        directory="keras_tuner"
    )

    tuner.search(
        x=X_train,
        y=y_train,
        epochs=max_epochs,
        batch_size=batch_size,
        validation_data=(X_valid, y_valid)
    )

    tuner.results_summary()
    # tuner.search_space_summary()
    winner = tuner.get_best_models(num_models=1)[0]
    winner.build(input_shape=[X_train.shape[1]])

    return winner



# https://towardsdatascience.com/looking-beyond-feature-importance-37d2807aaaa7
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from sklearn import metrics
from sklearn.metrics import confusion_matrix
import time


def feature_importance(model, X):
    n_features = X.shape[1]
    plt.barh(range(n_features), model.feature_importances_, align='center')
    plt.yticks(np.arange(n_features), X.keys())
    plt.xlabel("importance")
    plt.ylabel("feature")
    plt.show()


def chart(results_arr):
    plt.plot(list(results_arr.keys()), list(results_arr.values()))
    plt.show()


# more: https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html
# https://community.ibm.com/community/user/cloudpakfordata/blogs/harris-yang1/2021/05/26/scikit-learn-churn-model-cpd35
# https://www.kaggle.com/kanncaa1/roc-curve-with-k-fold-cv
def simple_roc(y_true, y_pred):
    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_pred, pos_label=1)
    plt.plot(fpr, tpr)
    plt.xlabel('False Positive Rate', fontsize=16)
    plt.ylabel('True Positive Rate', fontsize=16)
    plt.title('ROC', fontsize=18)


def simple_confusion_matrix(y_test, y_pred, labels):
    cm = confusion_matrix(y_test, y_pred, labels=labels)
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.matshow(cm, cmap=plt.cm.Blues, alpha=0.3)
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(x=j, y=i, s=cm[i, j], va='center', ha='center')
    plt.xlabel('Predicted label')
    plt.ylabel('True label')
    plt.tight_layout()
    plt.show()

    return cm


def simple_heatmap(df):
    data_correlations = df.corr()
    plt.subplots(figsize=(12, 10))
    sns.heatmap(data_correlations, cmap='Blues', annot=True)


def simple_correlations(df, target_name):
    cols = df.columns
    for label in cols:
        if label != target_name:
            sns.catplot(x=label, y=target_name, data=df)


def simple_features_overview(df):
    for label in df.columns:
        sns.catplot(x=label, data=df, height=3, aspect=4)



from simpleplotter import simple_features_overview, simple_correlations
from simpleplotter import simple_heatmap


def print_analytics(df, target_col=None):
    print("Data Overview:")
    print(df.head())

    print()
    print("Data Info:")
    print(df.info())

    print()
    print("Data Info2:")
    print(df.describe())

    print()
    print("Missing data:")
    print(df.isnull().sum())

    print()
    print("Features:")
    simple_features_overview(df)

    print()
    print("Correlations:")
    if target_col is not None:
        simple_correlations(df, target_col)
    simple_heatmap(df)
    print_simple_correlations(df)


def print_simple_correlations(df, level=0.3):
    print("Correlations >= " + str(level) + ":")
    for a in range(len(df.corr().columns)):
        col_a = df.corr().columns[a]
        print()
        print(col_a + ":")
        for b in range(a):
            corr = df.corr().iloc[a, b]
            if abs(corr) >= level:
                col_b = df.corr().columns[b]
                print(" - " + col_b + ": {:.2f}" . format(corr))
    print("Done")




from cross_validation_score import preprocess, separate_target, split_columns, split_train_test_80
from simple_analyzer import print_analytics, print_simple_correlations
from simple_hyper_tuner import quick_gridsearchcv_overview, gridsearchcv_tuner, predict_print_valid_scores, valid_score_tuner, print_valid_scores
from simple_processing import fillna_val_cols, fillna_mean, fillna_val, dropna_rows, remove_columns, separate_target, split_train_test, categorical_numerical, categorize, categorize_train_valid_test, print_scores, get_model_name, normalize, normalize_cols
from simple_reducer import umap_reducer, umap_reducer3d
from simple_yahoo import get_from_yahoo
from simpleplotter import feature_importance, chart, simple_roc, simple_confusion_matrix, simple_heatmap, simple_correlations, simple_features_overview
from single_scoring import get_scoring



import umap
from babyplots import Babyplot


def umap_reducer():
    embedding = reducer.fit_transform(scaled_penguin_data)
    embedding.shape


def umap_reducer3d(df, random_state=42, n_components=3):
    reducer = umap.UMAP(random_state=42)
    embedding = reducer.fit_transform(df)
    reducer3d = umap.UMAP(random_state=random_state, n_components=n_components)
    embedding3d = reducer3d.fit_transform(df)
    bp = Babyplot()
    bp.add_plot(
        embedding3d.tolist(),
        "pointCloud",
        "categories",
        y.values.tolist(),
        {
            "colorScale": "Set2",
            "showLegend": True,
            "folded": True,
            "foldedEmbedding": embedding.tolist()
        })
    html = bp.save_as_html('./babyplot.html')


print("Done")



import datetime
import pandas as pd


def date2unix(strdate):
    date = datetime.datetime.strptime(strdate, '%Y-%m-%d')
    return str(int(date.timestamp()))


def get_from_yahoo(currency, startDate, endDate):
    # Wczytanie danych z https://finance.yahoo.com/quote/BTC-USD/history?p=BTC-USD
    return pd.read_csv('https://query1.finance.yahoo.com/v7/finance/download/' + currency
                       + '?period1=' + date2unix(startDate) + '&period2=' + date2unix(endDate)
                       + '&interval=1d&events=history&includeAdjustedClose=true')



from tensorflow.keras.layers import Dense, Activation, Flatten
from tensorflow.keras.models import Sequential
from tensorflow.keras.metrics import MeanAbsoluteError
from tensorflow.keras.callbacks import TensorBoard
import keras_tuner
import pandas as pd
import datetime


# early_stopping = EarlyStopping(
#     min_delta=0.001,  # minimium amount of change to count as an improvement
#     patience=7,  # how many epochs to wait before stopping
#     restore_best_weights=True
# )


class SimpleRegressor:
    def __init__(self, X_train, y_train, X_valid, y_valid):
        self.X_train = X_train
        self.X_valid = X_valid
        self.y_train = y_train
        self.y_valid = y_valid
        self.model = None
        self.winner = None
        self.units_min = None
        self.units_max = None
        self.units_step = None
        self.features_count = self.X_train.shape[1]
        self.layers_count = 3

    def hyperband_search(self, max_epochs=5, batch_size=128, units_min=130, units_max=160,
                         units_step=4, layers_count=3):
        self.units_min = units_min
        self.units_max = units_max
        self.units_step = units_step
        self.layers_count = layers_count

        # tuner = keras_tuner.RandomSearch(
        tuner = keras_tuner.Hyperband(
            hypermodel=self.build_fit_callback,
            objective=keras_tuner.Objective("val_mean_absolute_error", direction="min"),
            max_epochs=50,
            factor=3,
            seed=42,
            overwrite=True,
            directory="logs/keras_tuner"
        )
        tuner.search(
            x=self.X_train,
            y=self.y_train,
            epochs=max_epochs,
            batch_size=batch_size,
            validation_data=(self.X_valid, self.y_valid)
        )
        tuner.results_summary()
        winner = tuner.get_best_models(num_models=1)[0]
        winner.build(input_shape=[self.features_count])
        self.winner = winner
        return winner

    def build_fit_callback(self, hp):
        activation = 'relu'
        units_tuner = hp.Int("units", min_value=self.units_min, max_value=self.units_max, step=self.units_step)
        # layers_count_tuner = hp.Int("units", min_value=min, max_value=max, step=self.units_step)

        self.build_model(units=units_tuner, activation=activation, layers_count=self.layers_count)
        self.compile_fit(epochs=10, batch_size=128)
        return self.model

    def build_model(self, units, activation, layers_count):
        m = Sequential()
        m.add(Dense(units=units, activation=activation, input_shape=[self.features_count]))
        for i in range(layers_count-1):
            m.add(Dense(units=units, activation=activation))

        m.add(Dense(1))
        self.model = m

    def compile_fit(self, epochs=400, batch_size=128):
        log_dir = "logs/tensorboard/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = TensorBoard(log_dir=log_dir, histogram_freq=1)

        self.model.compile(
            optimizer="adam",
            loss="mean_absolute_error",
            metrics=[MeanAbsoluteError()],
        )

        history = self.model.fit(x=self.X_train, y=self.y_train,
                                 validation_data=(self.X_valid, self.y_valid)
                                 , batch_size=batch_size
                                 , epochs=epochs
                                 # , callbacks=[early_stopping]
                                 , use_multiprocessing=True
                                 # , verbose='2'
                                 , callbacks=[tensorboard_callback]
                                 )
        return pd.DataFrame(history.history)

    def get_model(self):
        return self.model

    # m.compile(
    #     optimizer="adam",
    #     loss="mean_absolute_error",
    #     metrics=[keras.metrics.MeanAbsoluteError()],
    # )
    # # return m
    #
    # history = model.fit(x=self.X_train, y=self.y_train,
    #                     validation_data=(self.X_valid, self.y_valid)
    #                     , batch_size=128
    #                     , epochs=400
    #                     # , callbacks=[early_stopping]
    #                     , use_multiprocessing=True
    #                     # , verbose='2'
    #                     )
    # return pd.DataFrame(history.history)



# https://medium.com/all-things-ai/in-depth-parameter-tuning-for-gradient-boosting-3363992e9bae
import sys

from simple_processing import get_model_name, categorize, categorize_train_valid_test
from simpleplotter import simple_roc
from sklearn import svm
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier
import matplotlib.pyplot as plt


# LinearRegression (LR)
# RandomForest (RF)
# Support Vector Machine (SVM)


def quick_gridsearchcv_overview(X, y, cv=None, train_size=None, random_state=None, verbose=None):

    if cv is None:
        cv = [2]
    if random_state is None:
        random_state = [0]
    if train_size is None:
        train_size = [0.8]
    if verbose is None:
        verbose = 1

    params = [
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': GaussianNB(),
            'hyperparams': {}
        },
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': KNeighborsClassifier(
                # n_neighbors=19
            ),
            'hyperparams': {
                'n_neighbors': range(10, 25, 5)
            }
        },
        # {
        #     'splitter': {'train_size': train_size, 'random_state': random_state},
        #     'cv': cv,
        #     'model': LinearRegression(),
        #     'hyperparams': {}
        # },
        {   # Binary or Multiclass as One-Vs-One
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': svm.LinearSVC(max_iter=4000),
            'hyperparams': {}
        },
        {   # Binary or Multiclass - One-Vs-One and One-vs-Rest
            # https://www.baeldung.com/cs/svm-multiclass-classification
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': svm.SVC(),
            'hyperparams': {
                'decision_function_shape': ['ovr', 'ovo']
                , 'kernel': ['rbf', 'poly']
                # , 'shrinking': [True, False]
                # , 'coef0': [0, 0.2, 0.4]
                # , 'gamma' : ['scale', 'auto']
                # , 'degree': [1, 2, 3, 4, 5]
                # , 'C': [1, 0.1, 0.4, 0.7, 1.2]
            }
        },
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': LogisticRegression(max_iter=200, multi_class="ovr"),
            'hyperparams': {
                # 'tol': np.arange(0, 0.005, 0.001)
                # , 'C': [1, 0.1, 0.4, 0.7, 1.2]

            }
        },
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': DecisionTreeClassifier(random_state=0),
            'hyperparams': {
                # 'max_depth': [1, 2, 3, 4, 5],
                # 'max_leaf_nodes': range(2, 7, 1)
            }
        },
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': RandomForestClassifier(
                # max_depth=6, max_features="auto", bootstrap=True, oob_score=True, random_state=0
                # , criterion='entropy', n_estimators=75
            ),
            'hyperparams': {
                #         'max_features':["auto", "sqrt", "log2"],
                #         'max_samples': range(50, 80, 5),
                #         'n_estimators': range(40, 80, 10),
                #         'max_depth': range(4, 8, 1)
            }
        },
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': XGBClassifier(
                # n_estimators=70, learning_rate=0.08, max_depth=8, use_label_encoder=True
                # , num_parallel_tree=10
                # , min_samples_leaf=1, min_samples_split=3
                #   use_label_encoder=False, eval_metric='error'
                # , max_leaf_nodes=1, validation_fraction=0
                # , tol=0, ccp_alpha=0, n_iter_no_change=1
            ),
            'hyperparams': {
                # 'n_estimators':range(100, 160, 10)
                # 'min_samples_split':np.arange(3, 6, 1),
                # 'min_samples_leaf':np.arange(1, 4, 1),
                # 'subsample':np.arange(0.6, 0.9, 0.1),
                # 'criterion': ['friedman_mse', 'squared_error'],
                # , 'max_depth': range(5, 9, 1)
                # , 'learning_rate':np.arange(0.09, 0.13, 0.01)
                # , 'min_weight_fraction_leaf':np.arange(0, 0.14, 0.02)
                # , 'min_impurity_decrease':np.arange(0, 0.14, 0.02)
                # 'max_leaf_nodes':range(1, 7, 1)
                # , 'warm_start': [True, False]
                # 'validation_fraction': np.arange(0, 0.005, 0.001)
                # 'tol': np.arange(0, 0.005, 0.001)
                # 'ccp_alpha': np.arange(0, 0.005, 0.001)
                # 'n_iter_no_change':range(1, 7, 1)
            }
        },
        {
            'splitter': {'train_size': train_size, 'random_state': random_state},
            'cv': cv,
            'model': GradientBoostingClassifier(
                # loss='deviance', max_features="auto"
                # , criterion='squared_error', learning_rate=1.2, max_depth=6
                # , n_estimators=75, random_state=6
            ),
            'hyperparams': {
                #             'criterion': ['friedman_mse', 'mse', 'mae'],
                #             # 'loss':['deviance', 'exponential'],
                #             'random_state': range(5, 8, 1),
                #             'n_estimators': range(60, 90, 5),
                #             'learning_rate': np.arange(1, 1.4, 0.1),
                #             'max_depth': range(4, 9, 1)
            }
        }
    ]

    return gridsearchcv_tuner(X, y, params, verbose)


def gridsearchcv_tuner(X, y, params, verbose=1, do_categorize=True, X_test=None):
    best_model = {}
    best_f1 = 0

    print("X shape: ", X.shape)

    for p in params:
        for train_size in p['splitter']['train_size']:
            for random_state in p['splitter']['random_state']:
                # if p['reducer']:
                #     print('> Reducer: none')
                #     X_reduced = X
                # else:
                #     reducer = p['reducer']
                #     print('> Reducer:', reducer)
                #     reducer_model = reducer.fit(X)
                #     X_reduced = reducer_model.transform(X)
                #     # X_reduced = reducer_model.embedding_
                #     print('- X reduced shape: ', X_reduced.shape)
                    # plt.scatter(reducer_model.embedding_[:, 0], reducer_model.embedding_[:, 1], s=5, c=y, cmap='Spectral')
                    # plt.title('Embedding of the training set by reducer', fontsize=24)

                X_train, X_valid, y_train, y_valid = train_test_split(X, y, train_size=train_size, random_state=random_state)
                if do_categorize is True:
                    if X_test is None:
                        X_train, X_valid = categorize(X_train, X_valid)
                    else:
                        X_train, X_valid, X_test = categorize_train_valid_test(X_train, X_valid, X_test)

                for cv in p['cv']:
                    print('> Model:', get_model_name(p['model']))

                    grid_model = GridSearchCV(p['model'], p['hyperparams'], cv=cv, n_jobs=None, verbose=0)
                    grid_model.fit(X_train, y_train)

                    f1 = predict_print_valid_scores(grid_model, X_valid, y_valid, verbose)

                    if verbose == 0:
                        print("- cv: ", cv)
                    if verbose > 0:
                        print("GridSearchCV Training Results:")
                        print("- Best Score: ", grid_model.best_score_)
                        print("Params:")
                        print('- cv: ', cv)
                        print("- Splitter Params: ", {'train_size': train_size, 'random_state': random_state})
                        print("- Model Params: ", p['model'])
                        print("- Best H-Params: ", grid_model.best_params_)
                    if verbose == 2:
                        print("- All Params: ")
                        print(grid_model.get_params())
                    print()
                    if f1 > best_f1:
                        best_f1 = f1
                        best_model = grid_model.best_estimator_

    print('Done.')
    return best_model


def predict_print_valid_scores(model, X_valid, y_valid, verbose=1):
    y_pred = model.predict(X_valid)
    if verbose > 0:
        print('- Accuracy score: ', accuracy_score(y_valid, y_pred, normalize=True))
        print('- Accuracy count: ', accuracy_score(y_valid, y_pred, normalize=False), '/', y_pred.shape[0])
        print('- Precision score: ', precision_score(y_valid, y_pred, average='weighted'))
        print('- Recall score: ', recall_score(y_valid, y_pred, average='weighted'))
    f1 = f1_score(y_valid, y_pred, average='weighted')
    print('- F1 score: ', f1)
    simple_roc(y_valid, y_pred)
    return f1


def valid_score_tuner(X_train, y_train, X_valid, y_valid, parameters):
    max_train = 0
    max_valid = 0
    max_params = {}
    max_model_name = ''
    max_model = {}
    i = 0
    train_results = []
    test_results = []

    for p in parameters:
        modelName = p['modelName']
        hyperParamNames = p['hyperParamNames']
        hyperParamValues = p['hyperParamValues']
        params = p['params']
        fits = len(hyperParamValues[0]) * len(hyperParamValues[1]) * len(hyperParamValues[2])
        print('Fits to be done: ', fits)
        short_name = modelName.__name__

        for r1 in hyperParamValues[0]:
            for r2 in hyperParamValues[1]:
                for r3 in hyperParamValues[2]:

                    i += 1
                    sys.stdout.write(str(i) + '.. ')

                    hp = {
                        hyperParamNames[0]: r1,
                        hyperParamNames[1]: r2,
                        hyperParamNames[2]: r3
                    }

                    model = modelName(**hp, **params)
                    model.fit(X_train, y_train)

                    train_score = model.score(X_train, y_train)
                    valid_score = model.score(X_valid, y_valid)

                    if valid_score > max_valid:
                        max_model_name = short_name
                        max_model = model
                        max_train = train_score
                        max_valid = valid_score
                        max_params = hp

                        #                 if valid_score >= min_valid_score:
                        print()
                        print(short_name
                              + ' => train data SCORE: {:.3f}'.format(train_score)
                              + ' val data SCORE: {:.3f}'.format(valid_score))
                        print('params: ', max_params)

                        # for binary classifiers only
                        # false_positive_rate, true_positive_rate, thresholds = roc_curve(y_train, train_pred)
                        # roc_auc = auc(false_positive_rate, true_positive_rate)
                        # train_results.append(roc_auc)
                        #
                        # false_positive_rate, true_positive_rate, thresholds = roc_curve(y_valid, y_pred)
                        # roc_auc = auc(false_positive_rate, true_positive_rate)
                        # test_results.append(roc_auc)

    print()
    print('WINNER: ' + max_model_name)
    print(' => train data score: {:.3f}'.format(max_train)
          + ' val data score: {:.3f}'.format(max_valid))
    print('BEST params: ', max_params)

    y_pred = max_model.predict(X_valid)
    print_valid_scores(y_valid, y_pred)

    return max_model


def print_valid_scores(y_valid, y_pred):
    print('Accuracy score: ', accuracy_score(y_valid, y_pred, normalize=True))
    print('Accuracy count: ', accuracy_score(y_valid, y_pred, normalize=False), '/', y_pred.shape[0])
    print('Precision score: ', precision_score(y_valid, y_pred, average='weighted'))
    print('Recall score: ', recall_score(y_valid, y_pred, average='weighted'))
    print('F1 score: ', f1_score(y_valid, y_pred, average='weighted'))
    print()


#%%



from setuptools import setup

setup(
    name='dslibs',
    version='0.0.7',
    description='data science related tools & helpers',
    url='https://github.com/cjkepinsky/libs',
    author='Chris J. Kepinsky',
    author_email='chris.kepinski@gmail.com',
    license='MIT',
    packages=['dslibs'],
    install_requires=[
        'babyplots',
        'matplotlib',
        'numpy',
        'pandas',
        'scipy',
        'seaborn',
        'sklearn',
        'statistics',
        'umap',
        'xgboost',
    ],
    zip_safe=False
)



import re

import pandas as pd
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler


def fillna_val_cols(df, cols, val):
    df.update(df[cols].fillna(val, inplace=True))
    df.describe()
    return df


def fillna_mean(df, col):
    df[col].fillna(df[col].mean(), inplace=True)
    return df


def fillna_val(df, col, val):
    df[col].fillna(val, inplace=True)
    return df


def dropna_rows(df, columnNames):
    return df.dropna(axis=0, subset=columnNames, inplace=False)


def remove_columns(data, columnNames):
    return data.drop(columnNames, axis='columns')


def separate_target(data, column_name):
    y = data[column_name]
    X = data.drop([column_name], axis=1)
    return X, y


def split_train_test(x_train, y_train, train_size=0.8, test_size=0.2, random_state=0):
    return train_test_split(x_train, y_train, train_size=train_size, test_size=test_size,
                            random_state=random_state)


def categorical_numerical(x):
    # Select Categorical & numerical columns
    categorical = [cname for cname in x.columns if x[cname].nunique() < 10 and
                   x[cname].dtype == "object"]
    numerical = [cname for cname in x.columns if x[cname].dtype in ['int64', 'float64']]

    return categorical, numerical


def categorize(train, valid):
    train = pd.get_dummies(train)
    valid = pd.get_dummies(valid)
    return train.align(valid, join='left', axis=1, fill_value=0)


def categorize_train_valid_test(X_train_full, X_valid_full, X_test_full):
    # replace categorical data with numerical
    X_train = pd.get_dummies(X_train_full)
    X_valid = pd.get_dummies(X_valid_full)
    X_test = pd.get_dummies(X_test_full)

    X_train, X_valid = X_train.align(X_valid, join='left', axis=1, fill_value=0)
    X_train, X_test = X_train.align(X_test, join='left', axis=1, fill_value=0)

    return X_train, X_valid, X_test


def print_scores(y_valid, y_pred):
    print('Accuracy score: ', accuracy_score(y_valid, y_pred, normalize=True))
    print('Accuracy count: ', accuracy_score(y_valid, y_pred, normalize=False), '/', y_pred.shape[0])
    print('Precision score: ', precision_score(y_valid, y_pred, average='weighted'))
    print('Recall score: ', recall_score(y_valid, y_pred, average='weighted'))
    print('F1 score: ', f1_score(y_valid, y_pred, average='weighted'))
    print()

#
# @dispatch(GridSearchCV, DataFrame, Series)
# def print_scores(grid_model, X_valid, y_valid):
#     y_pred = grid_model.predict(X_valid)
#     print(grid_model.best_params_)
#     print('Accuracy score: ', accuracy_score(y_valid, y_pred, normalize=True))
#     print('Accuracy count: ', accuracy_score(y_valid, y_pred, normalize=False), '/', y_pred.shape[0])
#     print('Precision score: ', precision_score(y_valid, y_pred, average='weighted'))
#     print('Recall score: ', recall_score(y_valid, y_pred, average='weighted'))
#     print('F1 score: ', f1_score(y_valid, y_pred, average='weighted'))
#     print()


# helper function to retrieve model name from model object
def get_model_name(trained_model_obj):
    reg = re.compile('([A-Za-z]+)\(')
    return reg.findall(trained_model_obj.__str__())[0]


# https://machinelearningmastery.com/standardscaler-and-minmaxscaler-transforms-in-python/
def normalize(df, scaler=MinMaxScaler()):
    cols = df.columns
    # scaler = MinMaxScaler()

    df = scaler.fit_transform(df)
    df = pd.DataFrame(df, columns=cols)

    return df


def normalize_cols(cols):
    # cols = df.columns
    scaler = MinMaxScaler()

    cols = scaler.fit_transform(cols)
    # df = pd.DataFrame(df, columns=cols)

    return cols



# def print_summary(y_test, y_pred):
#     res = summary(y_test, y_pred)
#     print('Test data count: ', res['data_count'])
#     print('accuracy count: ', res['accuracy_count'])
#     print('accuracy score: ', res['accuracy'])
#     print('precision score: ', res['precision'])
#     print('recall score: ', res['recall'])
#     print()

import re
import sys


# X_train, y_train, X_valid, y_valid, p['modelName'], p['params'], p['hyperParams']
def get_scoring(X_train, y_train, X_valid, y_valid, modelName, hyperParamNames, hyperParamValues, params=None):
    if params is None:
        params = {}
    max_train = 0
    max_valid = 0
    max_model = ''
    max_params = {}
    max_model_name = ''
    i = 0
    train_results = []
    test_results = []

    reg = re.compile('([A-Za-z]+)\(')
    fits = len(hyperParamValues[0]) * len(hyperParamValues[1]) * len(hyperParamValues[2])
    print('Fits to be done: ', fits)
    short_name = reg.findall(modelName.__str__())[0]

    for r1 in hyperParamValues[0]:
        for r2 in hyperParamValues[1]:
            for r3 in hyperParamValues[2]:
                i += 1
                sys.stdout.write(str(i) + '.. ')

                hp = {
                    hyperParamNames[0]: r1,
                    hyperParamNames[1]: r2,
                    hyperParamNames[2]: r3
                }

                model = modelName(**hp, **params)
                model.fit(X_train, y_train)

                train_score = model.score(X_train, y_train)
                valid_score = model.score(X_valid, y_valid)
                # train_pred = model.predict(X_train)

                if valid_score > max_valid:
                    max_model_name = short_name
                    max_model = model
                    max_train = train_score
                    max_valid = valid_score
                    max_params = hp

                    #                 if valid_score >= min_valid_score:
                    print()
                    print(short_name
                          + ' => train data: {:.3f}'.format(train_score)
                          + ' val data: {:.3f}'.format(valid_score))
                    print('params: ', max_params)

                    # for binary classifiers only
                    # false_positive_rate, true_positive_rate, thresholds = roc_curve(y_train, train_pred)
                    # roc_auc = auc(false_positive_rate, true_positive_rate)
                    # train_results.append(roc_auc)
                    #
                    # false_positive_rate, true_positive_rate, thresholds = roc_curve(y_valid, y_pred)
                    # roc_auc = auc(false_positive_rate, true_positive_rate)
                    # test_results.append(roc_auc)

    print()
    print('WINNER: ' + max_model_name)
    print(' => train data: {:.3f}'.format(max_train)
          + ' val data: {:.3f}'.format(max_valid))
    print('BEST params: ', max_params)
    # y_pred = max_model.predict(X_valid)
    # df = DataFrame(y_valid, y_pred)
    # print(df)

    # import matplotlib.pyplot as plt
    # from matplotlib.legend_handler import HandlerLine2D
    # line1, = plt.plot(hyperParamValues[1], train_results, 'b', label='Train AUC')
    # line2, = plt.plot(hyperParamValues[1], test_results, 'r', label='Test AUC')
    # plt.legend(handler_map={line1: HandlerLine2D(numpoints=2)})
    # plt.ylabel('AUC score')
    # plt.xlabel('learning rate')
    # plt.show()

    # return {model: max_model, max_train: max_train, max_valid: max_valid}
